#include <stdio.h>
#include <stdlib.h>
#include <sys/types.h>
#include <sys/socket.h>
#include <netinet/in.h>         
#include <arpa/inet.h>
#include <sys/time.h>
#include <time.h>
#include <string.h>
#include <strings.h>
#include <unistd.h>
#include <sys/select.h>

#include "mcast.h"
#include "sap.h"

#define SAP_ADDR	"224.2.127.254"
#define SAP_PORT	9875
#define SAP_MAX_SIZE	1024

#define SAP_TIMEOUT	3

#define BUFFSIZE	SAP_MAX_SIZE

struct sdp_info {
	char *service;
	char *host;
	char *proto;
	char *port;
};

struct sdp_info *parse_sdp(char *sdp, int len)
{
	char *pos;
	static struct sdp_info sdpinfo;

	/* RFC 2327
	 * v=0
	 * o=- 6dca 1 IN IP4 192.168.100.17:2000
	 * s=TV Das Erste
	 * t=0 0
	 * c=IN IP4 192.168.100.17/1
	 * m=video 2000 http 33
	 * a=tool:getstream
	 * a=type:broadcast
	 */

	bzero(&sdpinfo, sizeof(struct sdp_info));

	pos = sdp;
	while(*pos != 0 && (pos-sdp) < len) {
		if (*pos == 0x0d)
			*pos = 0;

		if (*pos == 0x0a) {
			*pos = 0;

			if (!strncasecmp("s=", sdp, 2)) {
				sdpinfo.service = sdp + 2;
			} else if (!strncasecmp("c=", sdp, 2)) {
				int poscnt = 0;

				sdp += 2;
				while (*sdp != 0) {
					if (poscnt == 2 && *sdp == '/') {
						*sdp = 0;
						break;
					}

					if (*sdp == ' ') {
						*sdp = 0;
						poscnt++;

						/* c=<network type> <address type> <connection address> */
						if (poscnt == 2)
							sdpinfo.host = sdp + 1;

						if (poscnt > 2)
							break;
					}
					sdp++;
				}
			} else if (!strncasecmp("m=", sdp, 2)) {
				int poscnt = 0;

				sdp += 2;
				while (*sdp != 0) {
					if (*sdp == ' ') {
						*sdp = 0;
						poscnt++;

						/* m=<media> <port> <transport> <fmt list> */
						if (poscnt == 1)
							sdpinfo.port = sdp + 1;

						if (poscnt == 2)
							sdpinfo.proto = sdp + 1;

						if (poscnt > 2)
							break;
					}
					sdp++;
				}
			}

			sdp = ++pos;
			continue;
		}
		pos++;
	}

	return &sdpinfo;
}


char *get_url_from_sap(char *service)
{
	struct timeval start, curr;
	struct ip_mreq mreq;
	unsigned char buffer[BUFFSIZE];
	struct in_addr sapinaddr;
	char *url = NULL;
	int fd;

	snprintf(buffer,BUFFSIZE,"udp://%s:%u", SAP_ADDR, SAP_PORT);

	fd = open_mcast(buffer);

	gettimeofday(&start, NULL);

	do {
		int recvd;
		fd_set rfds;
		struct timeval tv;
		int retval;
		int sap_version, sap_addrtype, sap_messagetype, sap_encrypted, sap_compressed;
		uint8_t sender_address[16]; /* This might be IPv6! */
		unsigned char auth_len;
		unsigned short msgid;
		unsigned char *sdp;
		struct sdp_info *sdpinfo;

		FD_ZERO(&rfds);
		FD_SET(fd, &rfds);

		tv.tv_sec = 0;
		tv.tv_usec = 100000;

		if ((retval = select(fd+1, &rfds, NULL, NULL, &tv)) == -1) {
			perror("select");
			break;
		}

		if (!retval) {
			gettimeofday(&curr, NULL);
			continue;
		}

		bzero(buffer, BUFFSIZE);
		if ((recvd = recv(fd, buffer, BUFFSIZE, 0)) < 1) {
			perror("recv");
			break;
		}

		gettimeofday(&curr, NULL);

		sap_version = (buffer[0] >> 5) & 0x7;
		sap_addrtype = (buffer[0] >> 4) & 0x1;
		sap_messagetype = (buffer[0] >> 2) & 0x1;
		sap_encrypted = (buffer[0] >> 1) & 0x1;
		sap_compressed = buffer[0] & 0x1;
		auth_len = buffer[1];
		msgid = buffer[2] << 8 | buffer[3];
		memcpy(sender_address, buffer+4, (sap_addrtype?16:4));
		sdp = buffer + 4 /* (sap_*, auth_len, msgid) */ + (sap_addrtype?16:4) + auth_len;

#ifdef DEBUG
		printf("\n");
		printf("SAP-Version: %d\n", sap_version);
		printf("Adresstyp: %s\n", (sap_addrtype?"IPv6":"IPv4"));
		printf("Messagetype: %s\n", (sap_messagetype?"Announcement":"Deletion"));
		printf("Encrypted: %d\n", sap_encrypted);
		printf("Compressed: %d\n", sap_compressed);
		printf("Authentication Length: %d\n", auth_len);
		printf("Sender: %u\n", *((in_addr_t*)sender_address));
		printf("Message Identifier Hash: %u\n", msgid);
#endif

#if 0 /* Getstream gets this wrong, see rfc2974 */
		if (sap_messagetype)
			continue; /* We are not interested in deletions */
#endif

		if (sap_encrypted || sap_compressed)
			continue;

		sdpinfo = parse_sdp(sdp, recvd-(sdp-buffer));

		if (sdpinfo->service && sdpinfo->proto && sdpinfo->port) {
			char hostbuf[INET6_ADDRSTRLEN];

			if (!sdpinfo->host) {

				 if (sap_addrtype) {
					 struct in6_addr in6addr;

					 memcpy(in6addr.in6_u.u6_addr8, sender_address, 16);
					 if (!(sdpinfo->host = (char*)inet_ntop(AF_INET6, &in6addr, hostbuf, INET6_ADDRSTRLEN))) {
						 perror("inet_ntop");
						 continue;
					 }
				 } else {
					 struct in_addr inaddr;

					 inaddr.s_addr = *((in_addr_t*)sender_address);
					 if (!(sdpinfo->host = (char*)inet_ntop(AF_INET, &inaddr, hostbuf, INET6_ADDRSTRLEN))) {
						 perror("inet_ntop");
						 continue;
					 }
				 }
			}

#ifdef DEBUG
			printf("%s -> %s://%s:%s\n", sdpinfo->service, sdpinfo->proto, sdpinfo->host, sdpinfo->port);
#endif

			if (strlen(service) < strlen(sdpinfo->service)) {
				sdpinfo->service += strlen(sdpinfo->service) - strlen(service);
			}

			if (!strncasecmp(service, sdpinfo->service, strlen(service))) {
				int len = strlen(sdpinfo->host)+strlen(sdpinfo->proto)+strlen(sdpinfo->port)+5;

				if (!(url = malloc(len))) {
					perror("malloc");
					break;
				}

				snprintf(url, len, "%s://%s:%s", sdpinfo->proto, sdpinfo->host, sdpinfo->port);
				url[len-1] = 0;
				break;
			}
		}

	} while(curr.tv_sec < start.tv_sec+SAP_TIMEOUT);

	if (!(inet_aton(SAP_ADDR, &sapinaddr))) {
		perror("inet_aton");
	}

	mreq.imr_multiaddr.s_addr = sapinaddr.s_addr;
	mreq.imr_interface.s_addr = INADDR_ANY;
	setsockopt (fd, IPPROTO_IP, IP_DROP_MEMBERSHIP, &mreq, sizeof(mreq));

	close(fd);

	return url;
}
